VERY_SMALL_NUMBER = 1e-10
MAXINT = 2 ** 31 - 1
N_TREES = 10

class Key():
    REGRET_VAL = "regret_validation"
    REGRET_TEST = "regret_test"
    COST = "cost"
    BUDGET = "budget"
    EVAL_TIME = "eval_time"
    FUNC_VALUE = "function_value"